open System.Xml;
open Nemerle.Collections;

namespace Sioux
{
    class Config
    {
        private defs  : Hashtable (string, string);
        private doc   : XmlDocument;
        private fname : string;
        
	private put_defaults () : void
        {
            defs.Add ("server/name", "Sioux HTTP Server 0.001");
            defs.Add ("server/root", "/var/www");
            defs.Add ("server/port", "80");

            defs.Add ("replies/invalid", "Invalid request.");
            defs.Add ("replies/not_found", "File not found.");
        }
        
        public this (s : option (string))
        {
            defs  <- Hashtable ();
            doc   <- XmlDocument ();
            fname <- match (s){
                         | None     => "/etc/sioux/httpd.conf";
                         | Some (f) => f;
                     };
            try {
                doc.Load (fname);
            }
            with {
                e : System.Exception => ()
            };

            put_defaults ();
        }

        private get_default (key : string) : string
        {
            match (defs.Get (key)){
                | None     => null;
                | Some (s) => s;
            }
        }
        
        public get (key : string) : string
        {
            try {
                def sect   = doc.SelectSingleNode ("config/" + key);
                def reader = XmlNodeReader (sect);
                def _      = reader.Read ();
                def _      = reader.Read ();

                reader.Value
            }
            with {
                e : System.Exception => get_default (key)
            }
        }
    }
}
	